
/*
  testAlgoClient
  Ron Heath October 26, 2023

  usage: testAlgoClient 1234
     where 1234 is the key to shared memory of type char[sharedMemSize], *shm_ptr = the first char

  if *shm_ptr contains 'r' for "ready":
   program will look for two formulas, shm_ptr contains "r,M+m+V+v,m*v+M*V"
    after the first comma will be the formula for mass.
    after the second comma will be the formula for velocity

  if *shm_ptr contains 'k':
    this program will clean up and exit

  This program will change *shm_ptr to 's' or 'f' for score (or fail). if 's',
    the whole string will be "s2" where 2 is the score
    if the score was 0, it will change *shm_ptr to 'f' for failed

  note that the calling program should first write the two formulas before setting
  the first char, *shm_ptr, to an 'r' in order to avoid this program from starting while the calling program
  is still writing the formulas, strcpy(shm_ptr,"x,M+m+V+v,m*v+M*V"); then *shm_ptr = 'r';

  gcc bin/testAlgoOneD_client.o -o bin/testAlgoClient -lmatheval -L/usr/lib/x86_64-linux-gnu
  was: gcc your_program.c -o your_program -lrt
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <matheval.h>

#include <unistd.h>
#include <sys/stat.h>
#include <assert.h>
#include <dirent.h>
#include <sys/ipc.h>

#include <sys/ipc.h>
#include <sys/shm.h>
#include <time.h>

#define sharedMemSize 1000
#define maxFormulaSize sharedMemSize/2-4
#define arraySize 1000


key_t key;

typedef struct
{
	char formulaM[maxFormulaSize];
	char formulaV[maxFormulaSize];
} formStruct;
formStruct formulas;
int getFormulas(char *inputStr);



struct arrayElement
{
	double mass;
	double velocity;
};
struct arrayElement *arrayA;
struct arrayElement *arrayB;
struct arrayElement *lastT;

char fixedNames[4][10];
char *fixedNamesPtr[10];
char **fixedNamesPtrs;
double varValues[10];			/* Variable values  */


double absDouble(double inDbl);
void faceAlgorithm(struct arrayElement *element1, struct arrayElement *element2, struct arrayElement *result, void *fm, void *fv);
void runAlgorithm(struct arrayElement *arrayT0, struct arrayElement *arrayT1, void *fm, void *fv);
int evaluateFunction(char *formulaM, char *formulaV);
long nTests;
int score;


int main(int argc, char *argv[])
{
 int num;
 int shm_id;
 char *shm_ptr;
 char *tmpChPtr;




 if(argc != 2)
 {
  printf("Usage: %s <number>\n", argv[0]);
  printf("where <number> is the shared memory key\n");
  return 1;
 }
 num = atoi(argv[1]); // Convert the command line argument to an integer
 key = (key_t) num;
 printf("\rmy key is: %d\n\r", num);
 fflush(stdout);  // Flush the output buffer to ensure the message is printed immediately
	shm_id = shmget(key, sharedMemSize, 0666);
 if(shm_id < 0)
 {
 	perror("Error: could not get shm_id with shmget()\n");
 	exit(1);
 }
 shm_ptr = shmat(shm_id, NULL, 0);
 if(shm_ptr == (char*)-1)
 {
 	perror("Error: could not get shm_id with shmget()\n");
 	exit(1);
 }

 strcpy(fixedNames[0], "m");
 strcpy(fixedNames[1], "M");
 strcpy(fixedNames[2], "v");
 strcpy(fixedNames[3], "V");
 fixedNamesPtrs = fixedNamesPtr;
 fixedNamesPtrs[0] = &fixedNames[0][0];
 fixedNamesPtrs[1] = &fixedNames[1][0];
 fixedNamesPtrs[2] = &fixedNames[2][0];
 fixedNamesPtrs[3] = &fixedNames[3][0];


 //allocate memory for the arrays
 arrayA = (struct arrayElement *)malloc(arraySize * sizeof(struct arrayElement));
 if(arrayA == NULL)
 {
  printf("*** Error *** client %d: Memory allocation failed. Exiting...\n", key);
 	shmdt(shm_ptr); //detach the shared memory pointer
  return 1;
 }
 arrayB = (struct arrayElement *)malloc(arraySize * sizeof(struct arrayElement));
 if(arrayB == NULL)
 {
 	free(arrayA);
  printf("*** Error *** client %d: Memory allocation failed. Exiting...\n", key);
 	shmdt(shm_ptr); //detach the shared memory pointer
  return 1;
 }

 nTests = 0;


 while(1) //------------------------------main loop
 {
 	switch(*shm_ptr)
 	{
 	 case 'k':
 	 	strcpy(shm_ptr, "exit\n");
 	 	printf("\rclient %d Done!, exiting\n", key);
 	 	fflush(stdout);  // Flush the output buffer to ensure the message is printed immediately
 	 	shmdt(shm_ptr); //detach the shared memory pointer
 	 	free(arrayA);
 	 	free(arrayB);
 	 	return 0;
 	 	break;
   case 'r': //formulas are ready for processing
   	getFormulas(shm_ptr);
				score = evaluateFunction(formulas.formulaM, formulas.formulaV);
   	if(score > 2)
   	{
   		tmpChPtr = shm_ptr;
   		++tmpChPtr;
				 sprintf(tmpChPtr, ",%d\n", score); //write score to shared memory
				 *shm_ptr = 's'; // send the "score ready" reply
   	}
   	else
   	{
				 *shm_ptr = 'f'; //write failed to shared memory
   	}
   	break;
   case 't':
				printf("*** nTest ***\n");
				score = 3;
  		tmpChPtr = shm_ptr;
  		++tmpChPtr;
			 sprintf(tmpChPtr, ",%d\n", score); //write score to shared memory
			 *shm_ptr = 's'; // send the "score ready" reply
   	break;
   case 'x': //formulas are not ready
   default:
   	break;
 	}
 }
}
//return score
int getFormulas(char *inputStr) //---------------------------------------
{
 char stringX[maxFormulaSize];
 char *tmpChPtrIn, *tmpChPtrOut;
 int i, done;
 unsigned int partNum;


 partNum = 0;
 done = 0;
 tmpChPtrOut = stringX;
 for(tmpChPtrIn = inputStr, i = 0; (done == 0)&&(i < maxFormulaSize); tmpChPtrIn++, i++)
 {
 	switch(*tmpChPtrIn)
 	{
	  case ',':
	  	++partNum;
	  	if(partNum == 1)
	  	{
	  		*tmpChPtrOut = '\0';
	  		tmpChPtrOut = formulas.formulaM;
	  	}
	  	if(partNum == 2)
	  	{
	  		*tmpChPtrOut = '\0';
	  		i = i - strlen(formulas.formulaM); //there could be maxFormulaSize chars for formulas.formulaV
	  		tmpChPtrOut = formulas.formulaV;
	  	}
	  	if(partNum > 2)
	  	{
	  		printf("\rclient %d Error, more than three comma seperated parts: %s\n", key, inputStr);
	  		return 1;
	  	}
	  	break;
	  case '\n':
	  case '\0':
  		*tmpChPtrOut = '\0';
	  	if(partNum != 2)
	  	{
	  		printf("\rclient %d Error, fewer than three comma seperated parts: %s\n", key, inputStr);
	  		return 1;
	  	}
	  	done = 1;
	  	break;
	  default:
	  	*tmpChPtrOut = *tmpChPtrIn;
	  	++tmpChPtrOut;
 	}
 }

 //printf("stringF: %s\n", stringF);
 //printf("stringV: %s\n", stringV);

 return 0;

}
//------------------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------
void runAlgorithm(struct arrayElement *arrayT0, struct arrayElement *arrayT1, void *fm, void *fv)
{
int i;
struct arrayElement *aT0_minus, *aT0_center, *aT0_plus;
struct arrayElement result[2];
struct arrayElement *resultMinus, *resultPlus;

aT0_minus = aT0_center = aT0_plus = &arrayT0[1];
--aT0_minus; //this is now = &arrayT0[0]
++aT0_plus; //this is now = &arrayT0[2]
++arrayT1; //arrayT1[0] will not be modified, this is same position as aT0_center, but in arrayT1

resultMinus = &result[0];
resultPlus =  &result[1];


//printf("got to runAlgorithm\n"); //********************* DEBUG ******************


for(i = 0; i < (arraySize - 1); i++)
{

	if((aT0_center->mass == aT0_minus->mass)&&
		 (aT0_center->velocity == aT0_minus->velocity)&&
		 (aT0_center->mass == aT0_plus->mass)&&
		 (aT0_center->velocity == aT0_plus->velocity)
			)
	{ // if center voxel == plus == minus, no need to do algorithm
		arrayT1->mass = aT0_center->mass;
		arrayT1->velocity = aT0_center->velocity;
	}
	else
	{
	 faceAlgorithm(aT0_center, aT0_plus, resultPlus, fm, fv);
	 faceAlgorithm(aT0_center, aT0_minus, resultMinus, fm, fv);

	 arrayT1->mass = aT0_center->mass + resultPlus->mass + resultMinus->mass;
	 if(arrayT1->mass < 0) arrayT1->mass = 0; //no negative mass
	 arrayT1->velocity = aT0_center->velocity - resultMinus->velocity + resultPlus->velocity;
	}

	++aT0_minus;
	++aT0_center;
	++aT0_plus;
	++arrayT1;
}
}
//-----------------------------------------------------------------------------
// return the change in mass and the change in velocity
void faceAlgorithm(struct arrayElement *element1, struct arrayElement *element2, struct arrayElement *result, void *fm, void *fv)
{
int count = 4;



//printf("got to faceAlgorithm\n"); //********************* DEBUG ******************


if((element1->mass == element2->mass)&&(element1->velocity == element2->velocity))
{
	result->mass = element1->mass;
	result->velocity = element1->velocity;
	return;
}
if((element1->mass == 0)&&(element2->mass == 0))
{
	result->mass = 0;
	result->velocity = 0;
	return;
}


varValues[0] = element1->mass; // m, mass on this side of face
varValues[1] = element2->mass; // M, mass on the other side of face
varValues[2] = element1->velocity; // v, velocity on this side of face
varValues[3] = element2->velocity; // V, velocity on the other side of face

result->mass = evaluator_evaluate(fm, count, fixedNamesPtrs, varValues);
result->velocity = evaluator_evaluate(fv, count, fixedNamesPtrs, varValues);

}

//-----------------------------------------------------------------------------
double absDouble(double inDbl)
{
if(inDbl < 0)
	return 0 - inDbl;
return inDbl;
}
//-----------------------------------------------------------------------------
// return score for how well the algorithms performed
int evaluateFunction(char *formulaM, char *formulaV)
{
	int i;
	struct arrayElement *arrayMid;
	struct arrayElement *arrayPlus1, *arrayMinus1;
	int score;

	//--------------------------- these are for GNU libmatheval library functions
 void *fm, *fv;		/* Evaluators for mass and velocity functions.  */
 char **names;			/* Function variables names. */
 int count;			/* Number of function variables. */
 float lowestCntrMass;

 // itterationsToDo should be 7.87E-16 / tickTime
#define itterationsToDo 100

//	printf("got to evaluateFunction\n"); //********************* DEBUG ******************


/*
 *hydrogen:
 *atomic weight 1.0080 +/- 0.0002
 *Van Der Walls radius 120pm    pm is pico-meter or 1*10^-12 meter
 *volume = 0.5236 * d^3 = 0.5236 * (240E-12)^3 = 7.2382464E-30 m^3
 *the mass is 1.007825032241 Da, so the density is 7.2382464E-30 / 1.007825032241 = 7.18204665339E-30
 *
 *root mean square charge radius of a proton is about 0.84–0.87 fm (fm is femto-meter = 10^−15 meter).
 *the mass of the proton is 1.007276466621 Daltons
 *volume = 0.5236 * d^3
 *volume = 0.5236 * (.87fm * 2)^3 = 5.268024 fm^3
 *the mass is 1.007276466621 Da, so the density is 5.268024E-15 / 1.007276466621 = 5.22996831017E-15
 *
 *so the ratio of center mass to ambient mass should be 5.22996831017E-15 : 7.18204665339E-30 or 1 : 1.37324859874E-15
 *
 *the start mass of all of the voxels accept the center one is (proton mass) * (density ratio) or
 *  1.007276466621 * 1.37324859874E-15 = 1.38324099633E-15
 *
 * note that the term voxel is used for clarity so that a section of the 1 dimensional array can be visualized as having two faces.
 * This is a test only. When an algorithm has passed this test it will be applied to a 3 dimensional array and "voxel" will make more sense
 *
 */
#define initialCenter 1.007276466621     //proton mass in daltons
#define initialAmbient 1.38324099633E-15 // mass of one ambient voxel if the proton mass was expanded to the Van Der Walls radius


struct arrayElement *arrayT0;
struct arrayElement *arrayT1;



//next iteration array A will be the result or t = t+2 array
unsigned char presArrayT0; //the present t0 array, if = 0 arrayA is t0 and arrayB is t+1


	arrayT0 = &arrayA[0];
	arrayT1 = &arrayB[0];
	lastT = arrayT0;
	arrayMid = &lastT[arraySize / 2];
	score = 0;
	for(i = 0; i < arraySize; i++, ++arrayT0, ++arrayT1) //initialize the arrays
	{
		if(i == (arraySize / 2))
		{
			arrayT0->mass = initialCenter; //if center voxel
			arrayT1->mass = initialCenter;
		}
		else
		{
			arrayT0->mass = initialAmbient; //all the other voxels
			arrayT1->mass = initialAmbient;
		}
		arrayT0->velocity = 0;
		arrayT1->velocity = 0;
 }
	presArrayT0 = 0;
	arrayT0 = &arrayA[0];
	arrayT1 = &arrayB[0];

 //printf("\rclient %d inside evaluateFunction() mass = %s, velocity = %s\n", key, formulaM, formulaV);

 //sleep(1);

 //--------------- create Evaluator for mass function
 fm = evaluator_create (formulaM);
 if(fm == NULL) return -1;
 evaluator_get_variables (fm, &names, &count);
 if(count != 4)
 {
  //printf ("**** ERROR **** fm count != 4, count = %d, function = %s\n", count, formulaM);
  evaluator_destroy (fm);
  return 0;
 }

 //--------------- create Evaluator for velocity function
 fv = evaluator_create (formulaV);
 if(fv == NULL)
 {
  evaluator_destroy (fm);
 	return -1;
 }
 evaluator_get_variables (fv, &names, &count);
 if(count != 4)
 {
  //printf ("**** ERROR **** fv count != 4, count = %d, function = %s\n", count, formulaV);
  evaluator_destroy (fm);
  evaluator_destroy (fv);
  return 0;
 }



	lowestCntrMass = initialCenter;
	for(i = 0; i < itterationsToDo; i++)
	{

  if(presArrayT0 == 0) //run algorithm on arrayA and put result in arrayB
  {
  	arrayT0 = &arrayA[0];
  	arrayT1 = &arrayB[0];
  	lastT = &arrayB[0];
  	runAlgorithm(arrayT0, arrayT1, fm, fv);
  	presArrayT0 = 1;
  }
  else //if presArrayT0 == 1, so run algorithm on arrayB and put result in arrayA
  {
  	arrayT0 = &arrayB[0];
  	arrayT1 = &arrayA[0];
  	lastT = &arrayA[0];
  	runAlgorithm(arrayT0, arrayT1, fm, fv);
  	presArrayT0 = 0;
  }

		arrayMid = arrayPlus1 = arrayMinus1 = &lastT[arraySize / 2];
		++arrayPlus1;
		--arrayMinus1;
	 score = 0;

		if(i == 5)
		{
   double tmpDbl_1, tmpDbl_2;
			//-------------------------------------------------------------------test 1
			// verify the the center mass has decreased and the plus1 and minus1 voxels have increased
			// this should be done to see if the center mass is expanding

			if(arrayMid->mass < initialCenter)
			{
			//	printf("Iteration 5: center mass decreased\n");
				++score;
			}
			if((arrayPlus1->mass > initialAmbient)&&(arrayMinus1->mass > initialAmbient))
			{
			//	printf("Iteration 5: plus1 and minus1 voxels increased in mass\n");
				++score;
			}
			tmpDbl_1 = arrayMid->mass - arrayPlus1->mass;
			tmpDbl_2 = arrayMid->mass - arrayMinus1->mass;
			if(tmpDbl_1 != tmpDbl_2)
			{
				score = 0;
			}
   //---------------------------------------------------------------end "test 1"

			if(score < 2)
			{
				evaluator_destroy (fm);
		  evaluator_destroy (fv);
				return score;
			}
		}
  if(arrayMid->mass < lowestCntrMass) // get lowest center mass
  	lowestCntrMass = arrayMid->mass;

		if(i > 15) // look for bounce
		{
			if((arrayPlus1->mass > arrayMid->mass)&&(arrayMinus1->mass > arrayMid->mass))
			{
				score = 3;
			}
			if(arrayMid->mass > lowestCntrMass) // look for bounce
			{
				++score;
				if(score > 3)
				{
				 printf("*** Passed !!!!! ***\n");
				 evaluator_destroy (fm);
		   evaluator_destroy (fv);
				 return score;
				}
			}
		}
   //---------------------------------------------------------------end "test 2"
	}

 evaluator_destroy (fm);
 evaluator_destroy (fv);

 return 0;
}
//----------------------------------------------------------------------------------

